/*
 *	$Source: /home/nlfm/Working/Frink/RCS/tcl.c,v $
 *	$Date: 2000/08/22 12:25:32 $
 *	$Revision: 2.0.1.2 $
 *
 *------------------------------------------------------------------------
 *   AUTHOR:  Lindsay Marshall <lindsay.marshall@newcastle.ac.uk>
 *------------------------------------------------------------------------
 *    Copyright 1994-2000 The University of Newcastle upon Tyne (see COPYRIGHT)
 *========================================================================
 *
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>

#include "frink.h"

#ifdef __bsdi
#include <sys/types.h>
#include <sys/malloc.h>
#else
#include <malloc.h>
#endif

#define noToken	(Token *) 0

static int inClass = 0;

static Token lbraceToken	= {LBRACE, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
static Token rbraceToken	= {RBRACE, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
static Token xcontToken		= {XCONT, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
static Token ostartToken	= {OSTART, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
static Token startToken		= {START, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
static Token contToken		= {CONT, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
static Token econtToken		= {ECONT, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
static Token emToken		= {EM, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
/* static Token enToken		= {EN, CNSTWORD, 0, char *) 0, 0, (Token *) 0 , (Token *) 0}; */
static Token nospToken		= {NOSP, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
static Token *olsToken		= &emToken;
static Token spToken		= {SP, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
static Token lbrackToken	= {LBRACK, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
static Token rbrackToken	= {RBRACK, CNSTWORD, 0, (char *) 0, 0, (Token *) 0, (Token *) 0};
static Token msgcatToken	= {CONST, CNSTWORD, 0, "::msgcat::mc", 12, (Token *) 0, (Token *) 0};

static Token dqStart		= {DQSTART, CNSTWORD, 0, (char *) 0, 0, noToken, noToken};
static Token dqEnd		= {DQEND, CNSTWORD, 0, (char *) 0, 0, noToken, noToken};
static Token thenToken		= {CONST, CNSTWORD, 0, "then", 4, noToken, noToken};
static Token procToken		= {CONST, CNSTWORD, 0, "proc", 4, noToken, noToken};
static Token elseToken		= {CONST, CNSTWORD, 0, "else", 4, noToken, noToken};
static Token elseifToken	= {CONST, CNSTWORD, 0, "elseif", 6, noToken, noToken};
static Token methodToken	= {CONST, CNSTWORD, 0, "method", 6, noToken, noToken};
static Token commonToken	= {CONST, CNSTWORD, 0, "common", 6, noToken, noToken};
static Token protToken		= {CONST, CNSTWORD, 0, "protected", 9, noToken, noToken};
static Token semiToken		= {SEMI, CNSTWORD, 0, ";", 1, noToken, noToken};
static Token evalToken		= {CONST, CNSTWORD, 0, "eval", 4, noToken, noToken};
static Token codeToken		= {CONST, CNSTWORD, 0, "code", 4, noToken, noToken};

List *retinfo = (List *) 0;

static List *forinfo = (List *) 0;

static int retOp = 0;

extern void setIndent();
extern void outdent();
extern int isVarToken(Token *);
extern int isSingleCall(Token *, char *);
extern int tokEqual(Token *, char *);

static int doTest(int x)
{
   return (pragma != NOCHECK && (heuristics & x));
}

static int tokIsLevel(Token *tp)
{
    char *cp;
    if (tp == noToken || tp->type != CONST || tp->text == (char *) 0)
    {
	return 0;
    }
    cp = tp->text;
    if (*cp == '#')
    {
	cp++;
    }
    while (*cp)
    {
	if (*cp < '0' || *cp > '9')
	{
	    return 0;
	}
	cp++;
    }
    return 1;
}

static int oneLine(Token *seq, int semis)
{
    while (seq != noToken)
    {
	if (seq->type == SCOMMENT || seq->type == NL || 
	  (semis && seq->type == SEMI))
	{
	    return 0;
	}
	seq = seq->next;
    }
    return 1;
}

static int checkSpecial(char *val)
{
    char ch;

    if (*val == '{') return 0;
    while ((ch = *val++))
    {
	switch (ch)
	{
	case '$' :
	case '[' :
	    return 0;
	}
    }
    return 1;
}

static int single(Token * tp)
{
    if (tp != noToken && tp->next == noToken)
    {
	switch (tp->type)
	{
	case CONST :
	    return checkSpecial(tp->text);
	case CONC :
	case CALL :
	case VAR :
	    return 1;
	default :
	    return 0;
	}
    }
    return 0;
}

static int constant(Token * tp)
{
    return (tp != noToken && tp->next == noToken
	&& tp->type == CONST && !checkSpecial(tp->text));
}

/*
 * If there are no more tokens, print a useful message to the user and
 * exit.
 */
void failIfNullToken(Token *token, char *part, char *command, int ln)
{
    char msg[1024];

    if (token == noToken)
    {
	if (ln == 0)
	{
	    sprintf(msg, "Missing \"%s\" part in call of %s", part, command);
	}
	else
	{
	    sprintf(msg, "Missing \"%s\" part in call of %s, starting line %d",
	      part, command, ln);
	}
	fail(token, msg);
    }
}

/*
 * If there are more tokens after this one, print a useful warning to
 * the user.
 */
void warnIfArgsAfter(Token *token, char *part, char *command)
{
    char msg[1024];

    if (token != 0 && token->next != 0)
    {
	if (token->next->type == SCOMMENT)
	{
	    output(token->next, 0);
	}
	else
	{
	    if (*part != '\0')
	    {
		sprintf(msg, "Extra arguments after \"%s\" part in call of %s",
		  part, command);
	    }
	    else
	    {
		sprintf(msg, "Extra arguments in call of `%s'",
		  command);
	    }
	    warn(token, msg);
	    warn(token->next, "Extra Token is ");
	}
    }
}

void warnFor(Token *token, char *cmd, char *txt)
{
    char msg[1024];

    sprintf(msg, txt, cmd);
    warn(token, msg);
}

static int constantString(Token * tp, int any)
{
    Token *sq;
    if (tp == noToken) return 0;
    switch (tp->type)
    {
    case CONST :
	return 1;
    
    case LIST :
	return any;

    case STRING :
	sq = tp->sequence;
	return (sq->next == 0 && sq->type == SPART);

    default :
	return 0;
    }
}

static void msgsave(Token *tp, int any)
{
    char filename[128];

    if (extract && constantString(tp, any))
    {
	if (msgfile == NULL)
	{
	    if ((msgfile = fopen(strcat(strncpy(filename, locale, 123), ".msg"),
	      "a")) == NULL)
	    {
	    	warn(noToken, "cannot open msg file");
		return;
	    }
	}
	switch (tp->type)
	{
	case CONST:
	case LIST:
	    fprintf(msgfile, "::msgcat::mcset %s \"%s\"\n", locale, tp->text);
	    break;

	case STRING:
	    fprintf(msgfile, "::msgcat::mcset %s \"%s\"\n", locale,
	      tp->sequence->text);
	    break;

	default: /* shut up the compiler warnings */
	    break;
	}
    }
}

void body(Token *bod, int addBraces)
{
    Input *idx;
    Token *bl;
    int lnr;
    extern void sprocess(Token *, int);

    switch (bod->type)
    {
    case CONST :
	if (bod->ckind == CNSTSTRING)
	{
	    idx = tokenise(bod, 0);
	    bl = accumulate(idx, 1);
	    untokenise(idx);

	    output(&dqStart, 1);
	    setIndent();
	    if ((lnr = (oneliner && oneLine(bl, 0))))
	    {
		output(olsToken, 0);
	    }
	    lprocess(bl, 1);
	    if (lnr) { output(olsToken, 0); }
	    outdent();
	    output(&dqEnd, 0);
	    bod->sequence = noToken;
	    break;
	}
	else if (bod->ckind == CNSTLIST)
	{
    case LIST :
	    idx = tokenise(bod, 0);
	    bl = accumulate(idx, 1);
	    untokenise(idx);
	    if (debrace && bl != noToken && bl->next == 0
	      && bl->type == CONST && !checkSpecial(bl->text))
	    {
		output(bl, 1);
		freeToken(bl);
	    }
	    else if (oneliner && oneLine(bl, 0))
	    {
		output(&lbraceToken, 1);
		setIndent();
		output(olsToken, 0);
		lprocess(bl, 1);
		outdent();
		output(olsToken, 0);
		output(&rbraceToken, 1);
	    }
	    else
	    {
		output(&lbraceToken, 0);
		setIndent();
		lprocess(bl, 1);
		outdent();
		output(&rbraceToken, 0);
	    }
	    break;
	}
	if (embrace)
	{
	    output(&lbraceToken, 1);
	    setIndent();
	    output((oneliner) ? olsToken : &startToken, 0);
	}
	output(bod , 1);
	if (embrace)
	{
	    outdent();
	    if (oneliner) { output(olsToken, 0); }
	    output(&rbraceToken, oneliner);
	}
	break;
	
    case STRING :
        if (!trystrings)
	{
	    output(bod, 0);
	}
	else
	{
	    output(&dqStart, 1);
	    setIndent();
	    if ((lnr = oneliner && oneLine(bod->sequence, 1)))
	    {
	        output(olsToken, 0);
	    }
	    sprocess(bod->sequence, 1);
	    if (lnr) { output(olsToken, 0); }
	    outdent();
	    output(&dqEnd, 0);
	    bod->sequence = noToken;
	}
	break;

    case VAR :
    case CONC :
	if (addBraces && embrace)
	{
	    output(&lbraceToken, 1);
	    setIndent();
	    output((oneliner) ? olsToken : &startToken, 0);
	}
	output(bod , 1);
	if (addBraces && embrace)
	{
	    outdent();
	    if (oneliner) { output(olsToken, 0); }
	    output(&rbraceToken, oneliner);
	}
	break;

    case SLIST :
	output(bod, 0);
	break;

    case CALL :
	lprocess(bod->sequence, 0);
	bod->sequence = noToken;
	break;

    case SCOMMENT:
	output(bod, 1);
	break;
    case SEMI:
    case NL:
	blank();
	break;

    default :
	fail(bod, "Block error");
    }
}

typedef enum flags_e {
    NOBRACE	= 001,
    ADDBRACES	= 002,
    SEMIS	= 004,
    PAREN	= 010
} PressFlags;

static void press(Token *v , PressFlags flags)
{
    Input *idx;
    Token *token = noToken, *lst = noToken;
    int parcnt = 0;

    switch (v->type)
    {
    case SLIST:
	output(v, 1);
	break;
    case LIST :
	idx = tokenise(v, flags & PAREN);
	for(;;)
	{
	    token = getToken(idx);
	    switch (token->type)
	    {
	    case ENDF :
		freeToken(token);
		goto done;
	    case NL :
	    case SEMI :
	    case SCOMMENT :
		if (flags & SEMIS)
		{
		    tokenPush(&lst, token);
		    break;
		}
	    case PRAGMA:
	    case COMMENT :
	    case SP :
		freeToken(token);
	    	break;
	    case LPAREN:
		parcnt += 1;
		tokenPush(&lst, token);
		break;
	    case RPAREN :
		if ((parcnt -= 1) < 0)
		{
		    warn(v, "Possible missing '('");
		}
		tokenPush(&lst, token);
		break;
	    default :
		tokenPush(&lst, token);
	    }
	    idx->lineStart = 0;
	}
done :
	untokenise(idx);
	if (parcnt > 0)
	{
	    warn(v, "Possible missing ')'");
	}
	if (((flags & NOBRACE) && debrace && single(lst)) || constant(lst))
	{
	    output(lst, 1);
	}
	else
	{
	    output(&lbraceToken, 1);
	    token = lst;
	    while (token != noToken)
	    {
		switch (token->type)
		{
		case SEMI :
		    if (flags && SEMIS)
		    {
			output(&semiToken, 1);
		    }
		    else
		    {
			output(&startToken, 1);
		    }
		    break;
		default :
		    output(token, 1);
		}
		token = token->next;
	    }
	    output(&rbraceToken, 1);
	}
	freeToken(lst);
	break;
	
    case SCOMMENT:
	output(v, 1);
	break;
	
    case SEMI :
    case NL :
	if (flags & SEMIS) { output(&semiToken, 1); }
    	break;

    default :
	if (embrace && (flags & ADDBRACES)) { output(&lbraceToken, 1); }
	output(v , 1);
	if (embrace && (flags & ADDBRACES)) { output(&rbraceToken, 1); }
    }
}

void etcetera(Token *cmd, int v)
{
    while (cmd != noToken)
    {
	output(cmd, v);
	cmd = cmd->next;
    }
}

void catbin(Token *tp)
{
    int sem, oln;
    Token *sols;

    if (tp != noToken)
    {
	switch (tp->type)
	{
	case LIST :
	    if (tp->length == 0)
	    {
		output(tp, 1); /* empty list - just output it */
		return;
	    }
	    break;
	case STRING :
	default :
	    if (!trystrings)
	    {
		output(tp, 1);
		return;
	    }
	    break;
	}
	oln = oneliner;
	oneliner = 1;
	sem = embrace;
	embrace = 0;
	sols = olsToken;
	olsToken = &nospToken;
	body(tp, 0);
	olsToken = sols;
	embrace = sem;
	oneliner = oln;
    }
}

static void handlescom(Token *tp)
{
    if (tp != noToken)
    {
	output(tp, 1);
    }
}

static Token *handleOpt(Token *arg)
{
    Token *ap = arg;

    arg = arg->next;
    if ((!nocommand || pragma == NOFORMAT) && tokEqual(ap, "-command"))
    {
	catbin(arg);
		
    }
    else if (tokEqual(ap, "-text") || tokEqual(ap, "-label"))
    {
	if (!isSingleCall(arg, "::msgcat::mc"))
	{
	    if (internat)
	    {
		output(&spToken, 1);
		output(&lbrackToken, 1);
		output(&msgcatToken, 1);
		output(arg, 1);
		output(&rbrackToken, 1);
	    }
	    else
	    {
		output(arg, 1);
	    }
	    msgsave(arg, 1);
	}
	else
	{
	    output(arg, 1);
	}
    }
    else
    {
	output(arg, 1);
    }
    return arg;
}

void makeCall (Token *prc, Token *arg)
{
    if (prc != noToken)
    {
    	output(&startToken, 0);
	if (tokEqual(prc, "}")) { warn(prc, "unmatched } found"); }
	output(prc, 1);
    }
    while (arg != noToken)
    {
	if (arg->type == CONST && arg->text[0] == '-' &&
	  arg->next != noToken)
	{
/*
 * Be careful here - watch out for the case where -option is used as an
 * access function rather than to set the command!!
 */
	    if (xf)
	    {
		output(&xcontToken, 1);
	    }
	    output(arg, 1);
	    arg = handleOpt(arg);
	}
	else
	{
	    output(arg, 1);
	}
	arg = arg->next;
    }
}

void doswitch(Token *cmd)
{
    Token *tp, *bod;
    Input *idx;
    int ln = cmd->lineNo;
    Token *inbrace = &lbraceToken;
    Token *outbrace = &rbraceToken;

    tp = cmd;
    while (tp != noToken && tp->type == CONST && tp->text[0] == '-')
    {
	output(tp , 1 );
	if (tp->length == 2 && tp->text[1] == '-')
	{
	    tp = tp->next;
	    break;
	}
	tp = tp->next;
    }
    failIfNullToken(tp, "string", "switch", ln);
    output(tp, 1);
    tp = tp->next;
    failIfNullToken(tp, "pattern", "switch", ln);
    if (tp->next != noToken && tp->next->type != SCOMMENT)
    { /* this the non-list format */
	if (switchIn) { setIndent(); }
	while(tp != noToken && tp->type != SCOMMENT)
	{
	    output(&contToken, 1 );
	    output(tp, 1);
	    tp = tp->next;
	    failIfNullToken(tp, "body", "switch", ln);
	    if (tokEqual(tp, "-"))
	    {
		output(tp, 1);
	    }
	    else
	    {
		setIndent();
		body(tp, 0);
		outdent();
	    }
	    tp = tp->next;
	}
	handlescom(tp);
	if (switchIn) { outdent(); }
    }
    else
    {
	switch (tp->type)
	{
	case CONST :
	    if (tp->ckind == CNSTSTRING)
	    {
		inbrace = &dqStart;
		outbrace = &dqEnd;
	    }
	case LIST :
	    idx = tokenise(tp, 0);
	    tp = bod = accumulate(idx , 0);
	    untokenise(idx );
	    output(inbrace, 0);
	    if (switchIn) { setIndent(); }
	    while (tp != noToken)
	    {
		if (tp->type == COMMENT || tp->type == PRAGMA) {
		    output(tp, 1 );
		}
		else
		{
		    output(&ostartToken, 1);
		    output(tp, 1 );
		    tp = tp->next;
		    failIfNullToken(tp, "body", "switch", ln);
		    if (tokEqual(tp, "-"))
		    {
			output(tp, 1);
		    }
		    else
		    {
			setIndent();
			body(tp, 1);
			outdent();
		    }
		}
		tp = tp->next;
	    }
	    if (switchIn) { outdent(); }
	    output(outbrace, 0);
	    freeToken(bod);
	    break;
	case STRING :
/*	    break; */
	default :
	    failIfNullToken(noToken, "body", "switch", ln);
	}
    }
}

void doif(Token *cmd)
{
    Token *tp, *then;
    int efl = ADDBRACES, ln = cmd->lineNo;

    failIfNullToken(cmd, "condition", "if", ln);
    press(cmd, ADDBRACES | PAREN);
    if (putThen) { output(&thenToken, 0); }
    then = cmd->next;
    if (tokEqual(then, "then")) { then = then->next; }
    failIfNullToken(then, "then", "if", ln);
    body(then, 0);
    tp = then->next;
    while(tokEqual(tp, "elseif"))
    {
	if (!nonlelsif) { output(&econtToken, 1); }
	output(&elseifToken, 0);
	tp = tp->next;
	failIfNullToken(tp, "condition", "elseif", ln);
	press(tp, efl | PAREN);
	tp = tp->next;
	failIfNullToken(tp, "body", "elseif", ln);
	body(tp, 0);
	tp = tp->next;;
    }
    if (tokEqual(tp, "else"))
    {
	if (putElse) { output(&elseToken, 0 ); }
	tp = tp->next;
	failIfNullToken(tp, "body", "else", ln);
	body(tp, 0);
    }
    else if (tp != noToken)
    {
	if (tp->type != SCOMMENT)
	{
	    if (putElse) { output(&elseToken, 0 ); }
	    body(tp, 0);
	    warnIfArgsAfter(tp, "else", "if");
	}
	else
	{
	    output(tp, 0);
	}
    }
}

void checkNSBody(Token *cmd, int ln)
{
    if (tokEqual(cmd, "eval"))
    {
	output(&evalToken, 0);
	cmd = cmd->next;
	failIfNullToken(cmd, "name", "namespace", ln);
	output(cmd, 0);
	cmd = cmd->next;
	failIfNullToken(cmd, "script", "namespace", ln);
	if (cmd->next != noToken)
	{
	    etcetera(cmd, 0);
	}
	else
	{
	    catbin(cmd);
	}
    }
    else if (tokEqual(cmd, "code"))
    {
	output(&codeToken, 0);
	cmd = cmd->next;
	failIfNullToken(cmd, "script", "namespace", ln);
	catbin(cmd);
    }
    else
    {
	etcetera(cmd, 0);
    } 
}

void donamespace(Token *cmd)
{
    int ln = cmd->lineNo;

    checkNSBody(cmd, ln);
}

void dointerp(Token *cmd)
{
    int ln = cmd->lineNo;

    if (tokEqual(cmd, "eval"))
    {
	output(&evalToken, 0);
	cmd = cmd->next;
	failIfNullToken(cmd, "path", "interp", ln);
	output(cmd, 0);
	cmd = cmd->next;
	failIfNullToken(cmd, "script", "interp", ln);
	if (cmd->next != noToken)
	{
	    output(cmd, 0);
	}
	else
	{
	    catbin(cmd);
	}
    }
    else
    {
	etcetera(cmd, 0);
    } 
}

void doProc(Token *tag, Token *cmd)
{
    int ln = cmd->lineNo;
    Procret *pr;

    if (tag == &methodToken && !inClass)
    {
	makeCall(noToken, cmd);
    }
    else
    {
	pr = (Procret *) malloc(sizeof(Procret));
	pr->result = 0;
	pr->returns = 0;
	pr->pname = cmd;
	lpush(&retinfo, pr);
	
	failIfNullToken(cmd, "name", tag->text, ln);
	output(cmd, 1);
	cmd = cmd->next;
	failIfNullToken(cmd, "args", tag->text, ln);
	press(cmd, NOBRACE | ADDBRACES);
	cmd = cmd->next;
	failIfNullToken(cmd, "body", tag->text, ln);
	body(cmd, 0);
	warnIfArgsAfter(cmd, "body", tag->text);
	
	if (doTest(HRETURN) && pragma != NOTREACHED)
	{
	    if (!retOp && pr->returns != 0 && pr->result)
	    {
	    	warn(cmd, "possible missing final return statement");
	    }
	}
	(void) lpop(&retinfo);
	free(pr);
    }
}

void doproc(Token *cmd) { doProc(&procToken, cmd); }

void domethod(Token *cmd) { doProc(&methodToken, cmd); }

void dodestructor(Token *cmd)
{
    if (!inClass)
    {
	makeCall(noToken, cmd);
    }
    else
    {
	failIfNullToken(cmd, "body", "destructor", 0);
	body(cmd, 0);
	warnIfArgsAfter(cmd, "body", "destructor");
    }
}

void doconstructor(Token *cmd)
{
    if (!inClass)
    {
	makeCall(noToken, cmd);
    }
    else
    {
	failIfNullToken(cmd, "args", "constructor", 0);
	press(cmd, NOBRACE | ADDBRACES);
	cmd = cmd->next;
	failIfNullToken(cmd, "body", "constructor", 0);
	body(cmd, 0);
	warnIfArgsAfter(cmd, "body", "constructor");
    }
}

void doreturn(Token *cmd)
{
    Procret *pr;

    if ((heuristics & HRETURN))
    {
	if (retinfo != (List *) 0)
	{
	    pr = (Procret *) retinfo->item;
	    if (cmd != noToken)
	    {
		if (pr->returns != 0 && !pr->result)
		{
	    	    warn(cmd, "inconsistent return of values in proc");
		}
		pr->result = 1;
	    }
	    else if (pr->returns != 0 && pr->result)
	    {
	    	warn(cmd, "inconsistent return of values in proc");
	    }
	    pr->returns += 1;
	}
	else
	{
	    warn(cmd, "return used outside a proc body");
	}
    }
    makeCall(noToken, cmd);
    retOp = 1;
}

void doerror(Token *cmd)
{
    makeCall(noToken, cmd);
    retOp = 1;
}

void dobind(Token *cmd)
{
    Token *np;
    if (!doBind || pragma == NOFORMAT)
    {
	makeCall(noToken,cmd);
    }
    else
    {
	failIfNullToken(cmd, "windowSpec", "bind", 0);
	output(cmd, 1);
	if ((np = cmd->next) != noToken)
	{
	    output(np, 1);
	    if ((np->next) != noToken)
	    {
		catbin(np->next);
		warnIfArgsAfter(np->next, "command", "bind");
	    }
	}
    }
}

void doitcl_class(Token *cmd)
{
    failIfNullToken(cmd, "className", "itcl_class", 0);
    output(cmd, 1);
    inClass += 1;
    cmd = cmd->next;
    failIfNullToken(cmd, "body", "itcl_class", 0);
    body(cmd, 0);
    warnIfArgsAfter(cmd, "body", "itcl_class");
    inClass += -1;
}

void docvar(Token *cmd, Token *prt)
{
    if (!inClass)
    {
	makeCall(noToken, cmd);
    }
    else
    {
	failIfNullToken(cmd, "varName", prt->text, 0);
	output(cmd, 1);
	if ((cmd = cmd->next) != noToken)
	{
	    press(cmd, NOBRACE | ADDBRACES);
	    warnIfArgsAfter(cmd, "init", prt->text);
	}
    }
}

void dopublic(Token *cmd)
{
    if (!inClass)
    {
	makeCall(noToken, cmd);
    }
    else
    {
	failIfNullToken(cmd, "varName", "public", 0);
	output(cmd, 1);
	if ((cmd = cmd->next) != noToken)
	{
	    press(cmd, NOBRACE | ADDBRACES);
	    if ((cmd = cmd->next) != noToken)
	    {
		body(cmd, 0);
		warnIfArgsAfter(cmd, "config", "public");
	    }
	}
    }
}

void doprotected(Token *cmd) { docvar(cmd, &protToken); }

void docommon(Token *cmd) { docvar(cmd, &commonToken); }

static void checkVar(Token *cmd, List *pt, char *nm)
{
    List *vt;

    if (pt != (List *) 0)
    {
	vt = (List *) pt->item;
	while (vt != (List *) 0)
	{
	    if (strcmp((char *) vt->item, nm) == 0)
	    {
		warn(cmd, "Possible nested reuse of foreach variable");
	    }
	    vt = vt->next;
	}
	checkVar(cmd, pt->next, nm);
    }
}

static void addForVars(Token *cmd)
{
    Input *file;
    Token *vr, *vp;

    if (doTest(HFOREACH))
    {
	switch (cmd->type)
	{
	case CONST:
	    lpush((List **) &forinfo->item, newString(cmd->text));
	    checkVar(cmd, forinfo->next, cmd->text);
	    break;

	case LIST:
	    file = tokenise(cmd, 0);
	    vp = vr = accumulate(file, 0);
	    while (vp != noToken)
	    {
		lpush((List **) &forinfo->item, newString(vp->text));
		checkVar(vp, forinfo->next, vp->text);
		vp = vp->next;
	    }
	    untokenise(file);
	    freeToken(vr);
	    break;

	default:
	    break;
	}
    }
}

void doforeach(Token *cmd)
{
    List *ft;

    failIfNullToken(cmd, "varName", "foreach", 0);
    lpush(&forinfo, (void *) 0);
    do
    {
	output(cmd, 0);
	addForVars(cmd);
	cmd = cmd->next;
	failIfNullToken(cmd, "list", "foreach", 0);
	press(cmd, NOBRACE);
	cmd = cmd->next;
    }
    while (cmd != noToken && cmd->next != noToken);
    failIfNullToken(cmd, "body", "foreach", 0);
    body(cmd, 0);
    ft = (List *) lpop(&forinfo);
    ldel(&ft, 1);
}

void doloop(Token *cmd )
{
    Token *tp;
    if (!tclX)
    {
	makeCall(noToken, cmd);
    }
    else
    {
	failIfNullToken(cmd, "var", "loop", 0);
	press(cmd, NOBRACE | ADDBRACES);		/* var */
	tp = cmd->next;
	failIfNullToken(tp, "first", "loop", 0);
	press(tp, ADDBRACES);			   /* first */
	tp = tp->next;
	failIfNullToken(tp, "limit", "loop", 0);
	press(tp, ADDBRACES);			   /* limit */

	tp = tp->next;
	failIfNullToken(tp, "body", "loop", 0);

	if (tp->next != noToken)
	{
	    press(tp, NOBRACE | ADDBRACES);		/* incr */
	    tp = tp->next;
	}
	body(tp, 0);			/* body */
	warnIfArgsAfter(tp, "body", "loop");
    }
} 
 
void doexpr(Token *cmd)
{
    if (!doExpr)
    {
	makeCall(noToken,cmd);
    }
    else
    {
	if (cmd == noToken)
	{
	    warn(cmd, "Missing expression body");
	}
	else if (cmd->next == noToken)
	{
	    if (cmd->type != LIST)
	    {
		warn(cmd, "expr body not braced.");
	    }
	    press(cmd, ADDBRACES | PAREN);
	}
	else
	{
	    warn(cmd, "expr body not braced.");
	    etcetera(cmd, 1);
	}
    }
}

void doBasic(List *lp, Token *hd, Token *cmd)
{
    List *sp;
    ParamData *pt;
    int fnd, paramCount, brkpar;
    char msg[1024];

    paramCount = 0;
    while (cmd != noToken && lp != (List *) 0)
    {
	pt = (ParamData *) lp->item;
	lp = lp->next;
	paramCount += 1;
	switch (pt->type)
	{
	case PVAR:
	    if (doTest(HVAR) && isVarToken(cmd))
	    {
		sprintf(msg, "Parameter %d of call on %s may not be a variable name",
		   paramCount, hd->text);
		warn(cmd, msg);
	    }
	    output(cmd, 1);
	    break;

	case PCODE:
	    body(cmd, 0);
	    break;

	case PSCRIPT:
	    catbin(cmd);
	    break;

	case PSQN:
	    press(cmd, NOBRACE | ADDBRACES | SEMIS);
	    break;

	case PEXPR:
	    brkpar = 0;
	    if ((sp = pt->values) != (List *) 0)
	    {
	       sscanf(sp->item, "%x", &brkpar);
	    }
	    press(cmd, brkpar);
	    break;

	case PBLTN:
	/* !!!!!!!!!!!!!! Should never get here!!!! */
	    fail(cmd, "Internal error");
	    break;

	case PLEVEL:
	    if (pt->optional)
	    {
		if (!tokIsLevel(cmd)) {
		    continue;
		}
	    }
	    else
	    {
		warn(cmd, "level number expected");
	    }
	    output(cmd, 0);
	    break;

	case PCMDS:
	    if (cmd->next != noToken)
	    {
		etcetera(cmd, 0);
	    }
	    else
	    {
		catbin(cmd);
	    }
	    cmd = (Token *) 0; /* eat up whole line */
	    continue;

	case PSINGLE :
	case PANY:
	case POPTION:
	case PFLAG:
	case PWINDOW:
	    if (doTest(HVALUE) && cmd->type == CONST
	      && (sp = pt->values) != (List *) 0)
	    {
		fnd = 0;
		while (sp != (List *) 0)
		{
		    if (!strcmp(sp->item, cmd->text))
		    {
			fnd = 1;
			break;
		    }
		    sp = sp->next;
		}
		if (!fnd)
		{
		    sprintf(msg, "\"%s\" is an incorrect value for parameter %d of %s call",
		      cmd->text, paramCount, hd->text);
		    warn(cmd, msg);
		}
	    }
	    output(cmd, 0);
	    break;

	case PARGS:
	case POPTIONS:
	    etcetera(cmd, 1);
	    cmd = (Token *) 0;
	    continue;
	}
	cmd = cmd->next;
    }
    while (lp != (List *) 0)
    {
	pt = (ParamData *) lp->item;
	if (!pt->optional)
	{
	    switch (pt->type)
	    {
	    case PSINGLE:
	    case PARGS:
	    case POPTIONS:
	        break;
	    default:
	        warnFor(hd, hd->text, "Call of %s with too few parameters");
	        break;
	    }
	    break;
	}
	lp = lp->next;
    }
    if (cmd != noToken)
    {
	if (cmd->type == SCOMMENT)
	{
	    output(cmd, 1);
	}
	else if (doTest(HNUMBER))
	{
	    warnFor(hd, hd->text, "Too many arguments for call of %s");
	}
    }
}

typedef void (*DoFunc)(Token *);

static DoFunc builtins[] =
{
    0,			/* 0 */
    dobind,		/* 1 */
    docommon,		/* 2 */
    doconstructor,	/* 3 */
    dodestructor,	/* 4 */
    doerror,		/* 5 */
    doexpr,		/* 6 */
    doforeach,		/* 7 */
    doif,		/* 8 */
    dointerp,		/* 9 */			
    doitcl_class,	/* 10 */
    doloop,		/* 11 */			
    domethod,		/* 12 */
    donamespace,	/* 13 */
    doproc,		/* 14 */
    doprotected,	/* 15 */
    dopublic,		/* 16 */
    doreturn,		/* 17 */
    doswitch, 		/* 18 */
};

static void doUser(Token *hd, Token *cmd, ConfigData *cpt)
{
    output(&startToken, 0);
    output(hd, 1);

    if (cpt->oprtn == 0)
    {
	doBasic(cpt->params, hd, cmd);
    }
    else
    {
	builtins[cpt->oprtn](cmd);
    }
}

int tclop(Token *hd, Token *line)
{
    List *cpt = config;

    retOp = 0;
    while (cpt != (List *) 0)
    {
	if (!strcmp(hd->text, ((ConfigData *) cpt->item)->name))
	{
	   doUser(hd, line, (ConfigData *) cpt->item);
	   return 1;
	}
	cpt = cpt->next;
    }
    return 0;
}
